class BaseSwitcherException(Exception):
    def __init__(self, **kwargs):
        self.name = self.__class__.__name__
        self.message = kwargs.get("message", self.name)
        self.status = kwargs.get("status", 0)
        self.http_status_code = kwargs.get("http_status_code", 500)
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"


class BadRequestException(BaseSwitcherException):
    def __init__(self, **kwargs):
        self.retry = kwargs.get("retry", False)
        super().__init__(**kwargs)

    def must_retry(self):
        return self.retry


class ProviderErrorException(BaseSwitcherException):
    def __init__(self, **kwargs):
        self.service_name = kwargs.get("service_name")
        self.provider_name = kwargs.get("provider_name")
        self.response = kwargs.get("response", {})
        super().__init__(**kwargs)


class BadProviderException(BaseSwitcherException):
    def __init__(self, **kwargs):
        self.switch_provider = kwargs.get("switch_provider", False)
        super().__init__(**kwargs)

    def must_switch_provider(self):
        return self.switch_provider


class ExitException(BaseSwitcherException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class ValidationResponseException(BaseSwitcherException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.validation = kwargs.get("key", {})
        self.message = f"{self.name}: key `{str(self.validation)}`"


class ResponseNotFoundException(BaseSwitcherException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.validation = kwargs.get("key", {})
        self.message = f"{self.name}: key `{str(self.validation)}`"



from urllib.parse import urlparse


class CommonHelper(object):
    @staticmethod
    def get_attr(obj, attr: str, default: str = None):
        if type(obj) is dict:
            if hasattr(obj, attr):
                return getattr(obj, attr)

        if type(obj) is list:
            if attr in obj:
                return obj[attr]

        return default

    @staticmethod
    def is_url(string):
        try:
            result = urlparse(string)
            return all([result.scheme, result.netloc])
        except Exception:
            return False

    @staticmethod
    def format_data(data, format):
        if not data:
            return None

        if isinstance(data, str):
            for key, value in format.items():
                if value is not None:
                    data = data.replace(f'{{{key}}}', str(value))
            return data
        elif isinstance(data, dict):
            for key, value in data.items():
                data[key] = CommonHelper.format_data(value, format)

            return data



import os
from json import load

from .models.provider_model import ProviderModel
from .ssm.ssm_client import SSMClient


class ProviderCollection(object):
    def __init__(self, driver_type: str, **kwargs):
        self.providers = []
        self.driver_type = driver_type
        self.social_type = kwargs.get('social_type', None)
        self.service_name = kwargs.get('service_name', None)
        self.retry_when_not_found = kwargs.get('retry_when_not_found', False)
        self.root_path = kwargs.get('root_path', None)
        self.boto3_session = kwargs.get('boto3_session', None)
        self.ssm_client = SSMClient(self.boto3_session)
        self._prepare()

    def get_providers(self) -> list:
        """Get providers from SSM Parameter Store

        Returns:
            list: List of providers
        """
        return self.providers

    def _prepare(self):
        """Prepare to get providers

        Raises:
            Exception: can not get providers
        """
        providers = self.ssm_client.get_providers(self.social_type)

        if not providers:
            raise Exception('Can not get providers')

        self._get_providers_info(providers)

    def _get_providers_info(self, providers) -> None:
        """Get providers info from template file

        Args:
            providers (list): List of provider get from SSM Parameter Store

        Raises:
            Exception: Not found template file
        """
        if self.driver_type == 'json' and self.root_path:
            self._load_providers_from_json(providers)
        if self.driver_type == 'ssm':
            self._load_providers_from_parameter_store(providers)

        if not self.providers:
            raise Exception('No provider found.')

    def _build_file_template_dir(self, provider_name: str) -> str:
        """Build file template dir from provider name

        Args:
            provider_name (str): Provider name

        Returns:
            str: File template dir
        """

        file_dir = f'{self.root_path}/core/templates/' \
                   f'{self.social_type}/third_party_api_constants/{provider_name}/{self.service_name}.json'

        if not os.path.exists(file_dir):
            print("- File dir: ", file_dir)
            return False

        return file_dir

    def _load_providers_from_json(self, providers):
        """Load providers from json file

        Args:
            providers (list): list of providers get from SSM Parameter Store
        """
        for provider in providers:
            provider_name = provider.get('name', None)
            if not provider_name:
                continue

            provider_name = provider_name.lower().replace('-', '_')
            file_dir = self._build_file_template_dir(
                provider_name=provider_name)
            if not file_dir:
                continue

            with open(file_dir) as f:
                try:
                    provider_info = load(f)
                except ValueError:
                    print(
                        'Can not load provider info from json file {}'.format(file_dir))
                    continue

                request = provider_info.get('request', None)
                response = provider_info.get('response', None)
                provider_data = ProviderModel(
                    request=request,
                    response=response,
                    retry_when_not_found=self.retry_when_not_found,
                    **provider,
                )

                self.providers.append(provider_data)

    def _load_providers_from_parameter_store(self, providers):
        """Load providers from pramameter store
        Args:
            providers (list): list of providers get from SSM Parameter Store
        """
        for provider in providers:
            provider_name = provider.get('name', None)
            if not provider_name:
                continue

            provider_name = provider_name.lower()
            template_path = self.build_template_path_ssm(
                provider_name=provider_name)
            try:
                provider_info = self.ssm_client.get_template(template_path)
                if not provider_info:
                    raise Exception("Not found templates")
                # provider_info = loads(templates)
            except Exception:
                print('Can not load provider info from parameter store {}'.format(
                    template_path))
                continue

            request = provider_info.get('request', None)
            response = provider_info.get('response', None)
            provider_data = ProviderModel(
                request=request,
                response=response,
                retry_when_not_found=self.retry_when_not_found,
                **provider,
            )
            self.providers.append(provider_data)

    def build_template_path_ssm(self, provider_name):
        return '/rapid_api_templates' + "/" + self.social_type + "/" + provider_name + "/" + self.service_name



from json import dumps, loads
from datetime import datetime

import boto3


class SSMClient(object):

    def __init__(self, boto3_session=None) -> None:
        print('Start init SSM client', datetime.now())
        if boto3_session is not None:
            self.client: boto3.client = boto3_session.client("ssm")
        else:
            self.client: boto3.client = boto3.client("ssm")
        print('End init SSM client', datetime.now())

    def get_providers(self, social_type):
        providers = self._get_parameter_by_path(
            name='/rapid_api_providers' + "/" + social_type)
        sorted_providers = sorted(
            providers, key=lambda x: x['priority'], reverse=False)
        enabled_providers = list(
            filter(lambda x: x['status'] is True, sorted_providers))
        return enabled_providers

    def set_providers(self, provider):
        parameter_name = '/rapid_api_providers' + "/" + \
                         provider.get('social_type', None) + "/" + provider.get('name', None) + "-" + provider.get('id',
                                                                                                                   None)
        parameter = self._get_parameter(name=parameter_name)

        if not parameter:
            return

        parameter_keys = list(provider.keys())

        for key in parameter_keys:
            val = provider.get(key, None)
            if val is not None:
                parameter[key] = val

        if provider and parameter:
            if parameter['quotas_remaining'] == 0:
                parameter['is_over_quota'] = True

            parameter['die_count'] = parameter['die_count'] + \
                provider.get('session_fail', 0)
            parameter['success_count'] = parameter['success_count'] + \
                provider.get('session_success', 0)

        return self._set_parameter(
            name=parameter_name,
            value=dumps(parameter),
        )

    def _get_parameter(self, name, with_decryption=True):
        parameter_raw = self.client.get_parameter(
            Name=name,
            WithDecryption=with_decryption,
        )

        return loads(parameter_raw['Parameter']['Value'])

    def _get_parameter_by_path(self, name, with_decryption=True):
        parameter_raw = self.client.get_parameters_by_path(
            Path=name,
            Recursive=True,
            WithDecryption=with_decryption,
        )

        parameters = parameter_raw['Parameters']
        return list(map(lambda x: loads(x['Value']), parameters))

    def _set_parameter(self, name, value, type='String', overwrite=True):
        return self.client.put_parameter(
            Name=name,
            Value=value,
            Type=type,
            Overwrite=overwrite,
        )

    def get_template(self, path):
        template = self._get_parameter(name=path)
        return template



class dotdictify(dict):
    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError('expected dict')

    def __setitem__(self, key, value):
        if '.' in key:
            myKey, restOfKey = key.split('.', 1)
            target = self.setdefault(myKey, dotdictify())
            if not isinstance(target, dotdictify):
                raise KeyError('cannot set "%s" in "%s" (%s)' % (restOfKey, myKey, repr(target)))
            target[restOfKey] = value
        else:
            if isinstance(value, dict) and not isinstance(value, dotdictify):
                value = dotdictify(value)
            dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        if '.' not in key:
            return dict.__getitem__(self, key)
        myKey, restOfKey = key.split('.', 1)

        target = dict.__getitem__(self, myKey)

        try:
            restOfKeyList = restOfKey.split(".", 1)
            firstRestOfKeyType = type(int(restOfKeyList[0]))
        except ValueError:
            firstRestOfKeyType = str

        if firstRestOfKeyType is int:
            return dotdictify(target[int(restOfKeyList[0])]).__getattr__(restOfKeyList[1])

        try:
            restOfKey = int(restOfKey)
        except ValueError:
            pass

        if not isinstance(target, dotdictify) and not isinstance(target, list):
            raise KeyError('cannot get "%s" in "%s" (%s)' % (restOfKey, myKey, repr(target)))
        return target[restOfKey]

    def __contains__(self, key):
        if '.' not in key:
            return dict.__contains__(self, key)
        myKey, restOfKey = key.split('.', 1)
        target = dict.__getitem__(self, myKey)
        if not isinstance(target, dotdictify):
            return False
        return restOfKey in target

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default
        return self[key]

    __setattr__ = __setitem__
    __getattr__ = __getitem__



import base64
from pathlib import Path
from urllib.parse import urlparse


class Common:

    @staticmethod
    def get_project_root() -> Path:
        return Path(__file__).parents[5]

    @staticmethod
    def transform_dict_with_mapping(dict_, mapping, default_value=None):
        transformed_dict = {k: dict_.get(v, default_value) for k, v in mapping.items()}
        return transformed_dict

    @staticmethod
    def validate_schema(data, schema):
        error = {}
        try:
            schema().load(data)
        except Exception as e:
            error = e
        return data, error

    @staticmethod
    def is_url(url: str) -> bool:
        """
        Check input string is URL or not

        Input:
            - url (str): String need to validate

        Output: result after checked (True/False)
        """
        try:
            result = urlparse(url)
            return all([result.scheme, result.netloc])
        except ValueError:
            return False

    @staticmethod
    def hash_message(message_value: str) -> str:
        """
        Hash message into Base64.
        Input:
        - message_value (str): Message need to hash

        Output:
        - base64_message (str): Messsage after hashed

        """
        message_bytes = message_value.encode('utf-8')
        base64_bytes = base64.b64encode(message_bytes)
        base64_message = base64_bytes.decode('utf-8')
        return base64_message



from provider_switch.helpers.mapping_data import MappingData
from provider_switch.models.provider_model import ProviderModel


class ResponseHelper(object):
    def __init__(self, provider: ProviderModel):
        self.provider = provider
        provider_response = provider.response if provider.response else {}
        self.mapping_data_fields = provider_response.get('mapping_data_fields', {})
        self.default_response = dict({
            "data": {},
            "paging": {},
            "message": "",
            "description": "",
        })

    def standardize_response(self, response: dict) -> dict:
        mapped_response = {}
        if self.mapping_data_fields:
            mapped_response = MappingData().map(template=self.mapping_data_fields, input=response)
        else:
            mapped_response = response

        return mapped_response



from datetime import datetime

from provider_switch.exceptions import ResponseNotFoundException, ValidationResponseException
from provider_switch.helpers.common_helper import CommonHelper
from provider_switch.helpers.dotdictify import dotdictify
# from provider_switch.ssm.ssm_client import SSMClient


class ProviderModel(object):
    _excludes = ['_format_values', 'ssm_client', 'status', 'url', 'headers', 'request', 'response',
                 'retry_when_not_found']

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', None)
        self.social_type = kwargs.get('social_type', None)
        self.api_key = kwargs.get('api_key', None)
        self.die_count = kwargs.get('die_count', 0)
        self.success_count = kwargs.get('success_count', 0)
        self.status = kwargs.get('status', True)
        self.config = kwargs.get('config', {})
        self.priority = kwargs.get('priority', 0)

        self.is_over_quota = kwargs.get('is_over_quota', False)
        self.quotas = kwargs.get('quotas', 1000)
        self.current_usage_quotas = kwargs.get('current_usage_quotas', 0)
        self.quota_refresh_time = kwargs.get('quota_refresh_time', datetime.now().timestamp())

        self.session_fail = 0
        self.session_success = 0
        self.retry_delay = kwargs.get('retry_delay', 2)
        self.retry_when_not_found = kwargs.get('retry_when_not_found', False)
        self.retry_time = kwargs.get('retry_time', 1)
        self.timeout = kwargs.get('timeout', 30)

        self.request = kwargs.get('request', {})
        self.response = kwargs.get('response', {})

        self._format_values = dict(api_name=self.name, api_key=self.api_key)

        # self.ssm_client = SSMClient()

        headers = self.get_attr('config.default_headers')
        spec_headers = self.get_attr('request.specs.headers')
        headers.update(spec_headers) if spec_headers else headers

        path = self.get_attr('request.specs.path')

        if 'https://' not in path:
            path = "https://{api_name}.p.rapidapi.com" + path

        self.url = CommonHelper.format_data(path, self._format_values)
        self.headers = CommonHelper.format_data(headers, self._format_values)

    def _validation(self, format, data, is_reverse=True):  # noqa: C901
        format_keys = format.keys()
        data_dict = dotdictify(data)

        for key in format_keys:
            try:
                value = getattr(data_dict, key)
            except Exception:
                if is_reverse is True:
                    raise Exception('Key not found: {}'.format(key))
                else:
                    return True

            format_value = format.get(key, None)

            if type(format_value) == bool:

                if is_reverse is True:
                    if value != format_value:
                        raise ValidationResponseException(key=key)
                else:
                    if value == format_value:
                        raise ResponseNotFoundException(key=key)
                    
            elif type(format_value) == int:
                if format_value == value:
                    raise ResponseNotFoundException(key=key)

            elif type(format_value) == str:

                if is_reverse is True:
                    if format_value == '' and value is None:
                        raise ValidationResponseException(key=key)
                else:
                    if format_value == value:
                        raise ResponseNotFoundException(key=key)

                if format_value.__contains__('contain(\''):
                    format_value = format_value.replace('contain(\'', '').replace('\')', '')

                    if is_reverse is True:
                        if format_value not in value:
                            raise ValidationResponseException(key=key)
                    else:
                        if format_value in value:
                            raise ResponseNotFoundException(key=key)

    def check_response(self, response):
        try:
            format = self.response.get('format', None)
            success_format = format.get('success', None)
            not_found_format = format.get('not_found', None)

            if not success_format or not not_found_format:
                return {
                    "status": True,
                    "retry": False,
                }

            if not_found_format:
                self._validation(not_found_format, response, False)

            if success_format:
                self._validation(success_format, response)

            return {
                "status": True,
                "retry": False,
            }
        except ValidationResponseException:
            return {
                "status": False,
                "retry": True,
            }
        except ResponseNotFoundException:
            if self.retry_when_not_found:
                return {
                    "status": False,
                    "retry": True,
                }
            return {
                "status": False,
                "retry": False,
            }

    def get_attr(self, key) -> str:
        try:
            if not key:
                return None

            dicts = dotdictify(self.__dict__)
            return getattr(dicts, key)
        except Exception:
            return None

    def is_good(self):
        return self.status is True and self.is_over_quota is False

    def increment_session_fail(self):
        self.session_fail += 1

    def increment_session_success(self):
        self.session_success += 1

    def refresh_session_fail(self):
        self.session_fail = 0

    def set_inactive(self):
        self.status = False
        # self.ssm_client.set_providers(self.to_json())
        pass

    def set_active(self):
        self.status = True

    def sync(self, provider_data):
        # self.ssm_client.set_providers(provider={**self.to_json(), **provider_data})
        pass

    def to_json(self):
        self_dict = self.__dict__
        self_dict_keys = self_dict.keys()
        json_data = {}

        for key in self_dict_keys:
            if key not in self._excludes:
                json_data[key] = self_dict.get(key)

        return json_data



from datetime import datetime
from http import HTTPStatus
from pathlib import Path
from time import sleep

import requests

from provider_switch.constants.core_constants import Constant
from provider_switch.exceptions import (
    BadProviderException,
    BadRequestException,
    ExitException,
    ProviderErrorException,
)
from provider_switch.helpers.common_helper import CommonHelper
from provider_switch.helpers.response_helper import ResponseHelper
from provider_switch.models.provider_model import ProviderModel
from provider_switch.provider_collection import ProviderCollection
from provider_switch.utils.notify_service import NotifyService
from json import dumps


class ProviderSwitchHandler(object):
    service_name = None
    allowed_content_type = ["application/json", "application/json; charset=utf-8"]

    def __init__(self, service_name: str, social_type: str, **kwargs):
        self.service_name = service_name
        self.social_type = social_type
        self.allowed_providers = kwargs.get("allowed_providers", None)
        self.is_pagination = kwargs.get("is_pagination", False)
        self.num_of_pages = kwargs.get("num_of_pages", 1)
        self.retry_when_not_found = kwargs.get("retry_when_not_found", False)
        self.providers = []
        self.current_provider = None
        self.driver_type = kwargs.get("driver_type", "ssm")
        self.root_path = kwargs.get("root_path", None)
        self.boto3_session = kwargs.get("boto3_session", None)
        self.notify_service = NotifyService(
            channels=[
                {
                    "type": "webhook",
                    "webhook_url": "https://chat.googleapis.com/v1/spaces/AAAAoFde3Ho/messages?key=AIzaSyDdI0hCZtE6vySjMm-WEfRq3CPzqKqqsHI&token=xGDDaCfjzmbRPTdGU2gjjtmLb-bffbm5JamLTNFrlI8",
                }
            ],
            timestamp=False,
        )
        self._prepare()

    def _prepare(self) -> None:
        # print(Path(__file__))
        # print('Start prepare', datetime.now())
        providers = ProviderCollection(
            driver_type=self.driver_type,
            social_type=self.social_type,
            service_name=self.service_name,
            retry_when_not_found=self.retry_when_not_found,
            root_path=self.root_path,
            boto3_session=self.boto3_session,
        ).get_providers()
        if self.allowed_providers is not None:
            providers = [
                provider
                for provider in providers
                if provider.name in self.allowed_providers
            ]

        self.providers = providers
        # print('End  prepare', datetime.now())

    def _get_best_provider(self) -> ProviderModel:
        """Get best provider

        Raises:
            Exception: _description_

        Returns:
            ProviderModel: _description_
        """
        for provider in self.providers:
            if provider.is_good():
                return provider

        raise Exception("No provider available")

    def _request(
        self,
        method: str,
        url: str,
        headers: dict,
        params: dict,
        timeout: int = None,
        body: dict = None,
    ) -> dict:  # noqa: C901
        request_specs = self.current_provider.get_attr("request.specs")
        has_pagination = False

        pagination = request_specs.get("pagination", None)

        if pagination is not None:
            cursor_field = pagination.get("cursor_field", None)
            cursor_append_to = pagination.get("cursor_append_to", None)
            cursor_default_value = pagination.get("cursor_default_value", None)

            if (
                cursor_field is not None
                and cursor_append_to is not None
                and cursor_default_value is not None
            ):
                has_pagination = True

        if has_pagination is True:
            if cursor_append_to == "path":
                url = url + "/" + cursor_default_value

            if cursor_append_to == "param":
                params[cursor_field] = cursor_default_value
        try:
            response = getattr(requests, method)(
                url=url, headers=headers, params=params, timeout=timeout, json=body
            )
        except Exception as exc:
            if Constant.TIMEOUT_ERROR_TEXT in str(exc):
                raise BadRequestException(
                    status=0, http_status_code=500, message="Error", retry=True
                )
        status_code = response.status_code if response is not None else None
        response_headers = response.headers if response is not None else None
        content_type = (
            response_headers.get("content-type") if response_headers else None
        )
        provider_data = {}
        rate_limit_keys = self.current_provider.config.get("rate_limit_keys", None)

        # print('------------------')
        # print('Status code: ', status_code)
        # print('------------------')

        if rate_limit_keys is not None and response_headers is not None:
            rate_limit_quotas_key = rate_limit_keys.get("rate_limit_quotas_key", 0)
            rate_limit_remaining_key = rate_limit_keys.get(
                "rate_limit_remaining_key", 0
            )
            rate_limit_refresh_key = rate_limit_keys.get("rate_limit_refresh_key", 0)

            quotas = int(response_headers.get(rate_limit_quotas_key, 0))
            quotas_remaining = int(response_headers.get(rate_limit_remaining_key, 0))
            quota_refresh_time = int(response_headers.get(rate_limit_refresh_key, 0))

            if not quotas and not quotas_remaining and not quota_refresh_time:
                provider_data = {
                    "quotas": quotas,
                    "quotas_remaining": quotas_remaining,
                    "quota_refresh_time": quota_refresh_time,
                }

        encoded_content = response.content.decode("utf-8")
        if encoded_content == "":
            raise ProviderErrorException(
                status=0,
                http_status_code=status_code,
                service_name=self.service_name,
                provider_name=self.current_provider.name,
                response={"REQUEST_URL": response.request.url},
            )

        response_json = response.json()

        print("Status code: ", status_code)
        # print('Header: ', response_headers)
        # print('Body: ', body)

        if status_code == HTTPStatus.NOT_FOUND:
            raise ExitException(status=0, http_status_code=status_code)

        if (
            status_code != HTTPStatus.OK
            or content_type not in self.allowed_content_type
        ):
            raise BadRequestException(
                status=0,
                http_status_code=status_code,
                message=response_json.get("message") or "Error",
                retry=True,
                provider_data=provider_data,
            )

        validated = self.current_provider.check_response(response_json)

        if validated["status"] is False:
            if validated["retry"] is True:
                raise BadRequestException(
                    status=0,
                    http_status_code=500,
                    message="Error",
                    retry=True,
                    provider_data=provider_data,
                )
            else:
                raise ExitException(
                    status=0,
                    http_status_code=500,
                )

        self.current_provider.increment_session_success()
        # try:
        #     print('Sync parameter')
        #     # self.current_provider.sync(provider_data)
        # except Exception as e:
        #     print(f'Error sync parameter: {str(e)}')

        return ResponseHelper(self.current_provider).standardize_response(response_json)

    def _multiple_requests(
        self, method: str, url: str, headers: dict, params: dict, body: dict
    ) -> dict:
        """Multiple requests

        Args:
            variables (dict, optional): _description_. Defaults to None.

        Returns:
            dict: _description_
        """
        response = []

        for i in range(self.num_of_pages):
            response.append(
                self._request(
                    method=method, url=url, headers=headers, params=params, body=body
                )
            )

        return response

    def _make_request(self, variables: dict) -> dict:  # noqa: C901
        """Make request to RapidAPI

        Args:
            variables (dict): Variables

        Raises:
            BadProviderException: Bad provider exception
            BadRequestException: Bad request exception
            ExitException: Exit exception

        Returns:
            dict: Response
        """
        try:
            params = {}
            method = ""
            url = ""
            provider = self._get_best_provider()
            self.current_provider = provider
            request_specs = provider.get_attr("request.specs")

            url = provider.get_attr("url")
            method = request_specs.get("method", "get")
            req_headers = provider.get_attr("headers")
            session_fail = provider.get_attr("session_fail")
            timeout = (
                int(provider.get_attr("timeout"))
                if provider.get_attr("timeout")
                else None
            )

            if session_fail == provider.retry_time:
                raise BadProviderException(switch_provider=True)

            params = provider.get_attr("request.specs.params")
            json_payload = provider.get_attr("request.specs.body")

            # print('Provider: ', provider.name)
            print("URL: ", url)
            # print('Method: ', method)
            # print('Params: ', params)
            print("Variables: ", variables)

            if variables:
                url = CommonHelper.format_data(url, variables)

                if params:
                    for param in params:
                        params[param] = variables.get(params[param], None)

                if json_payload:
                    for json_payload_item in json_payload:
                        json_payload[json_payload_item] = variables.get(
                            json_payload[json_payload_item], None
                        )

            if self.is_pagination is True:
                return self._multiple_requests(
                    method=method,
                    url=url,
                    headers=req_headers,
                    params=params,
                    body=json_payload,
                )

            return self._request(
                method=method,
                url=url,
                headers=req_headers,
                params=params,
                timeout=timeout,
                body=json_payload,
            )
        except ProviderErrorException as e:
            self.notify_service.error(
                self.service_name,
                self.current_provider.name if self.current_provider else "UNKNOWN",
                dumps(
                    {
                        "EXCEPTION": "BadProviderException: must_retry",
                        "REQUEST": {
                            "method": method,
                            "url": url,
                            "params": params,
                        },
                        "RESPONSE": e.response,
                    }
                ),
            )
            print(
                "ProviderErrorException: ", e.service_name, e.provider_name, e.response
            )
            return {
                "error": True,
                "exception": "ProviderErrorException",
                "trace": {
                    "provider_name": e.provider_name,
                    "provider_name": e.provider_name,
                },
            }

        except BadRequestException as e:
            self.notify_service.error(
                self.service_name,
                self.current_provider.name if self.current_provider else "UNKNOWN",
                dumps(
                    {
                        "EXCEPTION": "BadProviderException: must_retry",
                        "REQUEST": {
                            "method": method,
                            "url": url,
                            "params": params,
                        },
                        "RESPONSE": str(e.message),
                    }
                ),
            )
            print("BadRequestException: ", e.message)
            if e.must_retry():
                print("BadRequestException must_retry")
                sleep(provider.retry_delay or 3)
                provider.increment_session_fail()
                return self._make_request(variables)
            else:
                return None

        except BadProviderException as e:
            self.notify_service.error(
                self.service_name,
                self.current_provider.name if self.current_provider else "UNKNOWN",
                dumps(
                    {
                        "EXCEPTION": "BadProviderException: must_switch_provider",
                        "REQUEST": {
                            "method": method,
                            "url": url,
                            "params": params,
                        },
                        "RESPONSE": str(e.message),
                    }
                ),
            )
            print("BadProviderException: ", e.message)
            if e.must_switch_provider():
                print("BadProviderException must_switch_provider")
                try:
                    provider.set_inactive()
                except Exception as e:
                    print(f"Error sync parameter: {str(e)}")
                return self._make_request(variables)
            else:
                return None

        except ExitException as e:
            self.notify_service.error(
                self.service_name,
                self.current_provider.name if self.current_provider else "UNKNOWN",
                dumps(
                    {
                        "EXCEPTION": "ExitException",
                        "REQUEST": {
                            "method": method,
                            "url": url,
                            "params": params,
                        },
                        "RESPONSE": str(e.message),
                    }
                ),
            )
            print("ExitException: ", e.message)
            return None

        except Exception as e:
            self.notify_service.error(
                self.service_name,
                self.current_provider.name if self.current_provider else "UNKNOWN",
                dumps(
                    {
                        "EXCEPTION": "CommonException",
                        "REQUEST": {
                            "method": method,
                            "url": url,
                            "params": params,
                        },
                        "RESPONSE": str(e),
                    }
                ),
            )
            print("Exception: ", e)
            return None

    def _refresh_providers(self):
        for i in range(len(self.providers)):
            self.providers[i].set_active()
            self.providers[i].refresh_session_fail()

    def get(self, variables: dict = None) -> dict:
        """GET request

        Args:
            variables (dict, optional): _description_. Defaults to None.

        Returns:
            dict: _description_
        """
        result = self._make_request(variables=variables or {})
        self._refresh_providers()
        return result



import re
from .dotdictify import dotdictify


class MappingData(object):

    def __init__(self) -> None:
        pass

    def map(self, input: dict = None, template: dict = None) -> dict:
        input = input if input else {}
        template = template if template else self.template

        if not input or not template:
            return {}

        mapped_data = {}
        if template:
            mapped_data = self._map_response(input, template)

        return {
            **input,
            **mapped_data
        }

    def _map_list(self, data: list, key: str) -> list:
        values = self.map(template=key, input=data)
        return values

    def _map_response(self, input: dict, template: dict) -> dict:  # noqa: C901
        mapped_data = {}
        input_vdict = dotdictify(input)
        regex = "\[.+?\]"  # noqa: W605

        for key, value in template.items():
            try:
                val_key = value
                map_key = key

                if type(value) is list:
                    val_key = '.'.join(value)

                if type(key) is str and re.search(regex, key):
                    key_raw = key.replace('[', '').replace(']', '')
                    keys = key_raw.split(':')
                    map_key = keys[0]
                    new_values = input_vdict.__getattr__(keys[1])
                    new_type = keys[2]

                    if new_type == 'list':
                        if type(new_values) is list:
                            data = []
                            for item in new_values:
                                data.append(self._map_list(item, val_key))

                            mapped_data[map_key] = data
                        else:
                            print('[Error] MappingData: ', keys[1], ' is not a list')
                    elif new_type == 'dict':
                        mapped_data[map_key] = self._map_list(new_values, val_key)
                    else:
                        continue
                elif type(key) is str and key == '*':
                    if ".*" in val_key:
                        temp_data = input_vdict.__getattr__(val_key.replace('.*', ''))
                    if val_key == "*":
                        temp_data = input_vdict
                    else:
                        temp_data = input_vdict.__getattr__(val_key)

                    mapped_data = {**mapped_data, **temp_data}
                    continue
                else:
                    if type(val_key) is str:
                        data = input_vdict.__getattr__(val_key)
                        mapped_data[map_key] = data
                    elif type(val_key) is dict:
                        mapped_data[map_key] = self._map_response(input, val_key)

            except Exception as e:
                # print('[Error] MappingData: ', e)
                continue

        try:
            make_response_data = dotdictify(mapped_data)
            return make_response_data
        except KeyError as e:
            print('[Error] MappingData: ', e)
            return {}



from datetime import datetime
from json import dumps

import requests


class NotifyService(object):
    def __init__(
        self,
        channels: list = None,
        timestamp: bool = True,
        debug=False,
        silent=False,
        env="development",
    ):
        self.debug = debug
        self.channels = channels
        self.timestamp = timestamp
        self.env = env
        self.silent = silent

    def error(self, name: str, type: str, event: str, channels=None):
        if channels is None:
            channels = ["webhook"]
        self.push(
            message="\n".join(
                [
                    f"🚨 Service name *{name}*",
                    f"- *Provider*: {type}",
                    f"- *Trace*: \n```{event}```",
                ]
            ),
            channels=channels,
        )

    def push(self, message: str, channels: list = None):
        if self.silent or not channels or len(self.channels) == 0:
            return
        for channel in self.channels:
            if channels is None or (
                channels is not None and channel.get("type", None) in channels
            ):
                if isinstance(message, list):
                    message = "\n".join(message)
                message = self._parse_message(
                    message=message, parse_mode=channel.get("parse_mode")
                )
                if (
                    channel.get("type") == "webhook"
                    and channel.get("webhook_url") is not None
                ):
                    self._send_webhook(url=channel.get("webhook_url"), message=message)
                elif (
                    channel.get("type") == "telegram"
                    and "access_token" in channel
                    and "chat_id" in channel
                ):
                    self._send_telegram_message(
                        message=message,
                        access_token=channel.get("access_token"),
                        chat_id=channel.get("chat_id"),
                    )

    def send_webhook(self, url: str, message: str, parse_mode: str = "google"):
        parsed_message = self._parse_message(message=message, parse_mode=parse_mode)
        self._send_webhook(url=url, message=parsed_message)

    @staticmethod
    def _send_telegram_message(
        message: str, access_token: str, chat_id: str, parse_mode="Markdown"
    ):
        url_tpl = f"https://api.telegram.org/bot{access_token}/sendMessage?chat_id={chat_id}&parse_mode={parse_mode}&text={message}"
        response = requests.post(
            url_tpl.format(
                {
                    "access_token": access_token,
                    "chat_id": chat_id,
                    "parse_mode": parse_mode,
                    "message": message,
                }
            )
        )
        if response.status_code == 200:
            print("📣 [telegram] sent!")

    def _parse_message(self, message: str, parse_mode: str):
        parsed_message = message

        # 1. Convert link
        if message.find("<") != -1 or message.find(">") != -1:
            custom_data = message[message.find("<") + 1 : message.find(">")]
            if (
                custom_data
                and custom_data.startswith("https://")
                and parse_mode != "google"
            ):
                url_list = custom_data.split("|")
                if len(url_list) == 2:
                    if parse_mode == "markdown":
                        parsed_message = message.replace(
                            f"<{custom_data}>", f"[{url_list[1]}]({url_list[0]})"
                        )

        timestamp = ""
        if self.timestamp:
            timestamp = f"[{datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')}]"

        return f"{timestamp} {parsed_message}"

    @staticmethod
    def _send_webhook(**kwargs):
        response = requests.post(
            kwargs.get("url", None),
            headers={"Content-Type": "application/json; charset=UTF-8"},
            json={
                "formattedText": kwargs.get("message"),
                "text": kwargs.get("message"),
            },
        )

        if response.status_code == 200:
            print("📣 Webhook - sent!")
